"""M2Crypto wrapper for OpenSSL X509 API.

Copyright (c) 1999-2004 Ng Pheng Siong. All rights reserved.

Portions created by Open Source Applications Foundation (OSAF) are
Copyright (C) 2004-2007 OSAF. All Rights Reserved.
Author: Heikki Toivonen
"""

import binascii
import logging

from M2Crypto import ASN1, BIO, EVP, m2, types as C
from typing import (
    Callable,
    List,
    Optional,
    Union,
    Iterator,
    Any,
    overload,
    Iterable,
)

FORMAT_DER = 0
FORMAT_PEM = 1

__g = globals()
for x in dir(m2):
    if x.startswith("X509_V_ERR"):
        __g[x] = getattr(m2, x)
    elif x.startswith("XN_FLAG_"):
        __g[x] = getattr(m2, x)
    elif x.startswith("X509_PURPOSE_"):
        __g[x] = getattr(m2, x)
    elif x.startswith("VERIFY_"):
        __g[x.lower()] = getattr(m2, x)

log = logging.getLogger(__name__)


class X509Error(ValueError):
    pass


m2.x509_init(X509Error)

V_OK: int = m2.X509_V_OK


class X509_Extension(object):
    """
    X509 Extension
    """

    def __init__(
        self, x509_ext_ptr: Optional[C.X509_EXTENSION] = None, _pyfree: int = 1
    ) -> None:
        self.x509_ext = x509_ext_ptr
        self._pyfree = _pyfree

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0) and self.x509_ext:
            m2.x509_extension_free(self.x509_ext)

    def _ptr(self) -> Optional[C.X509_EXTENSION]:
        return self.x509_ext

    def set_critical(self, critical: int = 1) -> int:
        """
        Mark this extension critical or noncritical. By default an
        extension is not critical.

        :param critical: Nonzero sets this extension as critical.
                         Calling this method without arguments will
                         set this extension to critical.
        :return: 1 for success, 0 for failure
        """
        if self.x509_ext is None:
            raise X509Error("Extension not initialized")
        return m2.x509_extension_set_critical(self.x509_ext, critical)

    def get_critical(self) -> int:
        """
        Return whether or not this is a critical extension.

        :return:  Nonzero if this is a critical extension.
        """
        if self.x509_ext is None:
            raise X509Error("Extension not initialized")
        return m2.x509_extension_get_critical(self.x509_ext)

    def get_name(self) -> str:
        """
        Get the extension name, for example 'subjectAltName'.
        """
        if self.x509_ext is None:
            raise X509Error("Extension not initialized")
        out = m2.x509_extension_get_name(self.x509_ext)
        return out.decode()

    def get_value(self, flag: int = 0, indent: int = 0) -> str:
        """
        Get the extension value, for example 'DNS:www.example.com'.

        :param flag:   Flag to control what and how to print.
        :param indent: How many spaces to print before actual value.
        """
        if self.x509_ext is None:
            raise X509Error("Extension not initialized")
        buf = BIO.MemoryBuffer()
        m2.x509_ext_print(buf.bio_ptr(), self.x509_ext, flag, indent)
        return (buf.read_all() or b"").decode()


def new_extension(
    name: str,
    value: str,
    critical: int = 0,
    _pyfree: int = 1,
    pkey: Optional[EVP.PKey] = None,
) -> X509_Extension:
    """
    Create a new X509_Extension instance using OpenSSL's internal V3 extension configuration.

    This method relies on the OpenSSL 'x509v3_ext_conf' function to parse a
    textual extension name and value into an X509_EXTENSION structure.

    :param name: The short name of the extension (e.g.,
                 'subjectAltName', 'basicConstraints').
    :param value: The value string for the extension (e.g.,
                  'DNS:example.com', 'CA:TRUE'). For complex
                  extensions like 'subjectKeyIdentifier', the
                  value often specifies the content, such as
                  'hash' to compute the hash of the subject's
                  public key.
    :param critical: Set to 1 to mark the extension as critical,
                  0 otherwise (default is 0).
    :param _pyfree: Internal flag (default 1).
    :param pkey: Optional EVP.PKey object. **Required** when the
                 extension value requires an associated public
                 key to be computed, such as when: - `name` is
                 'subjectKeyIdentifier' and `value` is 'hash'. If
                 an extension requires a public key context and
                 `pkey` is None, an `X509Error` will be raised to
                 prevent a Segmentation Fault (SIGSEGV) in modern
                 OpenSSL versions by accessing an uninitialized
                 context pointer.

    :return: A new `X509_Extension` instance.
    :raises X509Error: If the extension cannot be created (e.g.,
                 invalid name/value) or if a required context
                 (`pkey`) is missing.
    """
    ctx = m2.x509v3_set_nconf()

    # This block enforces context availability for subjectKeyIdentifier:hash
    # to prevent a SIGSEGV in OpenSSL > 3.0 when ctx->subject_cert is NULL.
    if name == "subjectKeyIdentifier" and value == "hash" and pkey is None:
        m2.x509v3_ctx_free(ctx)  # Clean up the context object before raising error
        raise X509Error(
            "Cannot create 'subjectKeyIdentifier:hash' without a public key (pkey) context."
        )

    if pkey is not None:
        m2.X509V3_CTX_set_nconf_pkey(ctx, pkey._ptr())
    x509_ext_ptr = m2.x509v3_ext_conf(None, ctx, name, str(value))
    if x509_ext_ptr is None:
        raise X509Error(
            "Cannot create X509_Extension with name '%s' and value '%s'" % (name, value)
        )
    x509_ext = X509_Extension(x509_ext_ptr, _pyfree)
    x509_ext.set_critical(critical)
    return x509_ext


class X509_Extension_Stack(object):
    """
    X509 Extension Stack

    :warning: Do not modify the underlying OpenSSL stack
              except through this interface, or use any OpenSSL
              functions that do so indirectly. Doing so will get the
              OpenSSL stack and the internal pystack of this class out
              of sync, leading to python memory leaks, exceptions or
              even python crashes!
    """

    def __init__(
        self, stack: Optional[C.STACK_OF_X509_EXTENSION] = None, _pyfree: int = 0
    ) -> None:
        self.pystack: List[X509_Extension] = []
        if stack is not None:
            self.stack = stack
            self._pyfree = _pyfree
            num = m2.sk_x509_extension_num(self.stack)
            for i in range(num):
                # Set _pyfree=0, the C stack owns the extension objects.
                self.pystack.append(
                    X509_Extension(
                        m2.sk_x509_extension_value(self.stack, i),
                        _pyfree=0,
                    )
                )
        else:
            self.stack = m2.sk_x509_extension_new_null()
            self._pyfree = 1

    def __del__(self) -> None:
        # see BIO.py - unbalanced __init__ / __del__
        if getattr(self, "_pyfree", 0):
            m2.sk_x509_extension_free(self.stack)

    def __len__(self) -> int:
        return len(self.pystack)

    def __getitem__(self, idx: int) -> X509_Extension:
        return self.pystack[idx]

    def __iter__(self) -> Iterator[X509_Extension]:
        return iter(self.pystack)

    def _ptr(self) -> C.STACK_OF_X509_EXTENSION:
        return self.stack

    def push(self, x509_ext: X509_Extension) -> int:
        """
        Push X509_Extension object onto the stack.

        :param x509_ext: X509_Extension object to be pushed onto the stack.
        :return: The number of extensions on the stack.
        """
        ext_ptr = x509_ext._ptr()
        if ext_ptr is None:
            raise X509Error("Cannot push an uninitialized extension")
        self.pystack.append(x509_ext)
        ret = m2.sk_x509_extension_push(self.stack, ext_ptr)
        assert ret == len(self.pystack)
        return ret

    def pop(self) -> Optional[X509_Extension]:
        """
        Pop X509_Extension object from the stack.

        :return: X509_Extension object that was popped, or None if there is
                 nothing to pop.
        """
        x509_ext_ptr = m2.sk_x509_extension_pop(self.stack)
        if x509_ext_ptr is None:
            assert len(self.pystack) == 0
            return None
        return self.pystack.pop()


class X509_Name_Entry(object):
    """
    X509 Name Entry
    """

    def __init__(self, x509_name_entry: C.X509_NAME_ENTRY, _pyfree: int = 0) -> None:
        """
        :param x509_name_entry: this should be OpenSSL X509_NAME_ENTRY binary
        :param _pyfree:
        """
        self.x509_name_entry = x509_name_entry
        self._pyfree = _pyfree

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.x509_name_entry_free(self.x509_name_entry)

    def _ptr(self) -> C.X509_NAME_ENTRY:
        return self.x509_name_entry

    def set_object(self, asn1obj: ASN1.ASN1_Object) -> int:
        """
        Sets the field name to asn1obj

        :param asn1obj:
        :return: 0 on failure, 1 on success
        """
        return m2.x509_name_entry_set_object(self.x509_name_entry, asn1obj._ptr())

    def set_data(self, data: bytes, type: int = ASN1.MBSTRING_ASC) -> int:
        """
        Sets the field value.

        :param data: data in a binary form to be set
        :return: 0 on failure, 1 on success
        """
        return m2.x509_name_entry_set_data(self.x509_name_entry, type, data)

    def get_object(self) -> ASN1.ASN1_Object:
        obj_ptr = m2.x509_name_entry_get_object(self.x509_name_entry)
        return ASN1.ASN1_Object(obj_ptr, _pyfree=0)

    def get_data(self) -> ASN1.ASN1_String:
        str_ptr = m2.x509_name_entry_get_data(self.x509_name_entry)
        return ASN1.ASN1_String(str_ptr, _pyfree=0)

    def create_by_txt(
        self, field: str, type: int, entry: bytes, length: int
    ) -> C.X509_NAME_ENTRY:
        """
        Creates and returns a new X509_NAME_ENTRY object.
        Note: This is a factory method that uses an existing instance's
        pointer space in a confusing way. It should likely be a static or
        class method for clarity. The corrected implementation below makes it work
        as written.
        """
        return m2.x509_name_entry_create_by_txt(
            self.x509_name_entry, field, type, entry, length
        )


class X509_Name(object):
    """
    X509 Name
    """

    nid = {
        "C": m2.NID_countryName,
        "SP": m2.NID_stateOrProvinceName,
        "ST": m2.NID_stateOrProvinceName,
        "stateOrProvinceName": m2.NID_stateOrProvinceName,
        "L": m2.NID_localityName,
        "localityName": m2.NID_localityName,
        "O": m2.NID_organizationName,
        "organizationName": m2.NID_organizationName,
        "OU": m2.NID_organizationalUnitName,
        "organizationUnitName": m2.NID_organizationalUnitName,
        "CN": m2.NID_commonName,
        "commonName": m2.NID_commonName,
        "Email": m2.NID_pkcs9_emailAddress,
        "emailAddress": m2.NID_pkcs9_emailAddress,
        "serialNumber": m2.NID_serialNumber,
        "SN": m2.NID_surname,
        "surname": m2.NID_surname,
        "GN": m2.NID_givenName,
        "givenName": m2.NID_givenName,
    }

    _alias_map = {
        "SP": "ST",
        "Email": "emailAddress",
    }

    def __init__(
        self, x509_name: Optional[C.X509_NAME] = None, _pyfree: int = 0
    ) -> None:
        """
        :param x509_name: this should be OpenSSL X509_NAME binary
        :param _pyfree:
        """
        if x509_name is not None:
            assert m2.x509_name_type_check(x509_name), "'x509_name' type error"
            self.x509_name = x509_name
            self._pyfree = _pyfree
        else:
            self.x509_name = m2.x509_name_new()
            self._pyfree = 1

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.x509_name_free(self.x509_name)

    def __str__(self) -> str:
        return m2.x509_name_oneline(self.x509_name)

    def __getattr__(self, attr: str) -> str:
        if attr in self.nid:
            assert m2.x509_name_type_check(self.x509_name), "'x509_name' type error"
            out = m2.x509_name_by_nid(self.x509_name, self.nid[attr])
            if out is None:
                raise AttributeError(attr)
            return out.decode()
        raise AttributeError(attr)

    def __setattr__(self, attr: str, value: Any) -> None:
        """
        :return: 1 for success of 0 if an error occurred.
        """
        if attr in self.nid:
            assert m2.x509_name_type_check(self.x509_name), "'x509_name' type error"
            self.add_entry_by_txt(attr, ASN1.MBSTRING_ASC, value, -1, -1, 0)
        else:
            super().__setattr__(attr, value)

    def __len__(self) -> int:
        return m2.x509_name_entry_count(self.x509_name)

    def __getitem__(self, idx: int) -> X509_Name_Entry:
        if not 0 <= idx < len(self):
            raise IndexError("index out of range")
        return X509_Name_Entry(m2.x509_name_get_entry(self.x509_name, idx), _pyfree=0)

    def __iter__(self) -> Iterator[X509_Name_Entry]:
        for i in range(len(self)):
            yield self[i]

    def _ptr(self) -> C.X509_NAME:
        return self.x509_name

    def add_entry_by_txt(
        self,
        field: str,
        type: int,
        entry: Union[str, bytes],
        len: int,
        loc: int,
        set: int,
    ) -> int:
        """
        Add X509_Name field whose name is identified by its name.

        :param field: name of the entry
        :param type: use MBSTRING_ASC or MBSTRING_UTF8
               (or standard ASN1 type like V_ASN1_IA5STRING)
        :param entry: value
        :param len: buf_len of the entry
               (-1 and the length is computed automagically)
        :param loc: determines the index where the new entry is
               inserted: if it is -1 it is appended.
        :param set: determines how the new type is added. If it is zero
               a new RDN is created.
               If set is -1 or 1 it is added to the previous or next RDN
               structure respectively. This will then be a multivalued
               RDN: since multivalues RDNs are very seldom used set is
               almost always set to zero.

        :return: 1 for success of 0 if an error occurred.
        """
        assert m2.x509_name_type_check(self.x509_name), "'x509_name' type error"

        # The SWIG wrapper for this function is quirky and appears to expect
        # Python strings for both field and entry, handling encoding internally.
        entry_str = entry.decode("utf-8") if isinstance(entry, bytes) else str(entry)

        canonical_field = self._alias_map.get(field, field)

        ret = m2.x509_name_add_entry_by_txt(
            self.x509_name, canonical_field, type, entry_str, len, loc, set
        )
        if not ret:
            err_code = m2.err_get_error()
            m2.err_clear_error()
            err_lib = m2.err_lib_error_string(err_code)
            err_func = m2.err_func_error_string(err_code)
            err_reason = m2.err_reason_error_string(err_code)
            raise X509Error(
                "Failed to set attribute '{}'. OpenSSL error: [{}] {}: {}".format(
                    field, err_lib, err_func, err_reason
                )
            )
        return ret

    def entry_count(self) -> int:
        return m2.x509_name_entry_count(self.x509_name)

    def get_entries_by_nid(self, nid: int) -> List[X509_Name_Entry]:
        """
        Retrieve the next index matching nid.

        :param nid: name of the entry (as m2.NID* constants)

        :return: list of X509_Name_Entry items
        """
        ret = []
        lastpos = -1

        while True:
            lastpos = m2.x509_name_get_index_by_nid(self.x509_name, nid, lastpos)
            if lastpos == -1:
                break

            ret.append(self[lastpos])

        return ret

    def as_text(self, indent: int = 0, flags: int = m2.XN_FLAG_COMPAT) -> str:
        """
        as_text returns the name as a string.

        :param indent: Each line in multiline format is indented
                       by this many spaces.
        :param flags:  Flags that control how the output should be formatted.
        """
        assert m2.x509_name_type_check(self.x509_name), "'x509_name' type error"
        buf = BIO.MemoryBuffer()
        m2.x509_name_print_ex(buf.bio_ptr(), self.x509_name, indent, flags)
        return (buf.read_all() or b"").decode()

    def as_der(self) -> bytes:
        assert m2.x509_name_type_check(self.x509_name), "'x509_name' type error"
        return m2.x509_name_get_der(self.x509_name)

    def as_hash(self) -> int:
        assert m2.x509_name_type_check(self.x509_name), "'x509_name' type error"
        return m2.x509_name_hash(self.x509_name)


class X509(object):
    """
    X.509 Certificate
    """

    def __init__(self, x509: Optional[C.X509] = None, _pyfree: int = 0) -> None:
        """
        :param x509: binary representation of
               the underlying OpenSSL X509 object.
        """
        if x509 is not None:
            assert m2.x509_type_check(x509), "'x509' type error"
            self.x509 = x509
            self._pyfree = _pyfree
        else:
            self.x509 = m2.x509_new()
            self._pyfree = 1

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.x509_free(self.x509)

    def _ptr(self) -> C.X509:
        return self.x509

    def as_text(self) -> str:
        assert m2.x509_type_check(self.x509), "'x509' type error"
        buf = BIO.MemoryBuffer()
        m2.x509_print(buf.bio_ptr(), self.x509)
        return (buf.read_all() or b"").decode()

    def as_der(self) -> bytes:
        assert m2.x509_type_check(self.x509), "'x509' type error"
        return m2.i2d_x509(self.x509)

    def as_pem(self) -> bytes:
        assert m2.x509_type_check(self.x509), "'x509' type error"
        buf = BIO.MemoryBuffer()
        m2.x509_write_pem(buf.bio_ptr(), self.x509)
        return buf.read() or b""

    def save_pem(self, filename: str) -> int:
        """
        :param filename: name of the file to be loaded
        :return: 1 for success or 0 for failure
        """
        with BIO.openfile(filename, "wb") as bio:
            return m2.x509_write_pem(bio.bio_ptr(), self.x509)

    def save(self, filename: str, format: int = FORMAT_PEM) -> int:
        """
        Saves X.509 certificate to a file. Default output
        format is PEM.

        :param filename: Name of the file the cert will be saved to.

        :param format: Controls what output format is used to save the cert.
                       Either FORMAT_PEM or FORMAT_DER to save in PEM or
                       DER format.  Raises a ValueError if an unknow
                       format is used.

        :return: 1 for success or 0 for failure
        """
        with BIO.openfile(filename, "wb") as bio:
            if format == FORMAT_PEM:
                return m2.x509_write_pem(bio.bio_ptr(), self.x509)
            elif format == FORMAT_DER:
                return m2.i2d_x509_bio(bio.bio_ptr(), self.x509)
            else:
                raise ValueError(
                    "Unknown filetype. Must be either FORMAT_PEM or FORMAT_DER"
                )

    def set_version(self, version: int) -> int:
        """
        Set version of the certificate.

        :param version: Version number.
        :return:        Returns 0 on failure.
        """
        return m2.x509_set_version(self.x509, version)

    def get_version(self) -> int:
        return m2.x509_get_version(self.x509)

    def set_not_before(self, asn1_time: ASN1.ASN1_TIME) -> int:
        """
        :return: 1 on success, 0 on failure
        """
        return m2.x509_set_not_before(self.x509, asn1_time._ptr())

    def set_not_after(self, asn1_time: ASN1.ASN1_TIME) -> int:
        """
        :return: 1 on success, 0 on failure
        """
        return m2.x509_set_not_after(self.x509, asn1_time._ptr())

    def get_not_before(self) -> ASN1.ASN1_TIME:
        time_ptr = m2.x509_get_not_before(self.x509)
        time = ASN1.ASN1_TIME(time_ptr, _pyfree=0)
        time.owner = self
        return time

    def get_not_after(self) -> ASN1.ASN1_TIME:
        time_ptr = m2.x509_get_not_after(self.x509)
        time = ASN1.ASN1_TIME(time_ptr, _pyfree=0)
        time.owner = self
        return time

    def get_serial_number(self) -> int:
        asn1_int = m2.x509_get_serial_number(self.x509)
        py_int = int(ASN1.ASN1_Integer(asn1_int))
        if py_int is None:
            raise X509Error("Could not retrieve serial number")
        return py_int

    def set_serial_number(self, serial: int) -> int:
        """
        Set serial number.

        :param serial:  Serial number.

        :return 1 for success and 0 for failure.
        """
        asn1_integer = ASN1.ASN1_Integer(serial)
        return m2.x509_set_serial_number(self.x509, asn1_integer.asn1int)

    def get_pubkey(self) -> EVP.PKey:
        pkey_ptr = m2.x509_get_pubkey(self.x509)
        if pkey_ptr is None:
            raise X509Error("Could not get public key")
        return EVP.PKey(pkey_ptr, _pyfree=1)

    def set_pubkey(self, pkey: EVP.PKey) -> int:
        """
        Set the public key for the certificate

        :param pkey: Public key

        :return 1 for success and 0 for failure
        """
        return m2.x509_set_pubkey(self.x509, pkey.pkey)

    def get_issuer(self) -> X509_Name:
        return X509_Name(m2.x509_get_issuer_name(self.x509), _pyfree=0)

    def set_issuer(self, name: X509_Name) -> int:
        return m2.x509_set_issuer_name(self.x509, name._ptr())

    set_issuer_name = set_issuer

    def get_subject(self) -> X509_Name:
        return X509_Name(m2.x509_get_subject_name(self.x509), _pyfree=0)

    def set_subject_name(self, name: X509_Name) -> int:
        """
        :return: 1 on success, 0 on failure
        """
        return m2.x509_set_subject_name(self.x509, name._ptr())

    set_subject = set_subject_name

    def add_ext(self, ext: X509_Extension) -> int:
        """
        Add X509 extension to this certificate.

        :param ext:    Extension

        :return 1 for success and 0 for failure
        """
        ext_ptr = ext._ptr()
        if ext_ptr is None:
            raise X509Error("Cannot add an uninitialized extension")
        return m2.x509_add_ext(self.x509, ext_ptr, -1)

    def get_ext_count(self) -> int:
        """
        Get X509 extension count.
        """
        return m2.x509_get_ext_count(self.x509)

    def get_ext_at(self, index: int) -> X509_Extension:
        """
        Get X509 extension by index.

        :param index:    Name of the extension

        :return:        X509_Extension
        """
        if not 0 <= index < self.get_ext_count():
            raise IndexError("index out of range")
        ext_ptr = m2.x509_get_ext(self.x509, index)
        return X509_Extension(ext_ptr, _pyfree=0)

    def get_ext(self, name: str) -> X509_Extension:
        """
        Get X509 extension by name.

        :param name:    Name of the extension

        :return:       X509_Extension
        """
        for i in range(self.get_ext_count()):
            ext = self.get_ext_at(i)
            if ext.get_name() == name:
                return ext
        raise LookupError

    def sign(self, pkey: EVP.PKey, md: str) -> int:
        """
        Sign the certificate.

        :param pkey: Public key

        :param md:   Message digest algorithm to use for signing,
                     for example 'sha1'.

        :return int
        """
        mda = getattr(m2, md, None)
        if mda is None:
            raise ValueError("unknown message digest", md)
        return m2.x509_sign(self.x509, pkey.pkey, mda())

    def verify(self, pkey: Optional[EVP.PKey] = None) -> int:
        if pkey:
            return m2.x509_verify(self.x509, pkey.pkey)
        else:
            pubkey = self.get_pubkey()
            return m2.x509_verify(self.x509, pubkey.pkey)

    def check_ca(self) -> int:
        """
        Check if the certificate is a Certificate Authority (CA) certificate.

        :return: 0 if the certificate is not CA, nonzero otherwise.

        :requires: OpenSSL 0.9.8 or newer
        """
        return m2.x509_check_ca(self.x509)

    def check_purpose(self, id: int, ca: int) -> int:
        """
        Check if the certificate's purpose matches the asked purpose.

        :param id: Purpose id. See X509_PURPOSE_* constants.

        :param ca: 1 if the certificate should be CA, 0 otherwise.

        :return: 0 if the certificate purpose does not match, nonzero
                 otherwise.
        """
        return m2.x509_check_purpose(self.x509, id, ca)

    def get_fingerprint(self, md: str = "sha1") -> str:
        """
        Get the fingerprint of the certificate.

        :param md: Message digest algorithm to use.

        :return:   String containing the fingerprint in hex format.
        """
        md_obj = EVP.MessageDigest(md)
        md_obj.update(self.as_der())
        digest = md_obj.final()
        return binascii.hexlify(digest).upper().decode()

    def add_subject_key_identifier(self) -> int:
        """
        Adds a Subject Key Identifier (SKI) extension to the certificate,
        calculating the key identifier from the certificate's public key.

        This bypasses the error-prone use of 'subjectKeyIdentifier', 'hash'
        in new_extension, which requires a pre-configured OpenSSL context.

        :return: 1 for success and 0 for failure.
        """
        pkey = self.get_pubkey()
        # The SKID value is typically the SHA1 hash of the public key's contents.
        # We compute the hash manually and then format it for the extension value.
        skid_digest = pkey.get_key_identifier()

        # Format the digest into the required OpenSSL extension value string format:
        # A list of hex octets separated by colons (e.g., 'A1:B2:C3:D4:...')
        value = ":".join(f"{b:02X}" for b in skid_digest)

        # Create the extension directly with the computed value
        # Note: SKID is usually non-critical (critical=0 by default for new_extension)
        ext = new_extension("subjectKeyIdentifier", value)

        # Add the extension to the certificate object
        return self.add_ext(ext)


def load_cert(file: Union[str, bytes], format: int = FORMAT_PEM) -> X509:
    """
    Load certificate from file.

    :param file: Name of file containing certificate in either DER or
                 PEM format.

    :param format: Describes the format of the file to be loaded,
                   either PEM or DER.

    :return: M2Crypto.X509.X509 object.
    """
    if isinstance(file, bytes):
        file = file.decode()
    with BIO.openfile(file) as bio:
        if format == FORMAT_PEM:
            return load_cert_bio(bio)
        elif format == FORMAT_DER:
            cptr = m2.d2i_x509(bio._ptr())
            return X509(cptr, _pyfree=1)
        else:
            raise ValueError("Unknown format. Must be either FORMAT_DER or FORMAT_PEM")


def load_cert_bio(bio: BIO.BIO, format: int = FORMAT_PEM) -> X509:
    """
    Load certificate from a bio.

    :param bio: BIO pointing at a certificate in either DER or PEM format.

    :param format: Describes the format of the cert to be loaded,
                   either PEM or DER (via constants FORMAT_PEM
                   and FORMAT_FORMAT_DER)

    :return: M2Crypto.X509.X509 object.
    """
    if format == FORMAT_PEM:
        cptr = m2.x509_read_pem(bio._ptr())
    elif format == FORMAT_DER:
        cptr = m2.d2i_x509(bio._ptr())
    else:
        raise ValueError("Unknown format. Must be either FORMAT_DER or FORMAT_PEM")
    if cptr is None:
        raise X509Error("Failed to load certificate from BIO")
    return X509(cptr, _pyfree=1)


def load_cert_string(string: Union[str, bytes], format: int = FORMAT_PEM) -> X509:
    """
    Load certificate from a cert_str.

    :param cert_str: String containing a certificate in either
                     DER or PEM format.

    :param format: Describes the format of the cert to be loaded,
                   either PEM or DER (via constants FORMAT_PEM
                   and FORMAT_FORMAT_DER)

    :return: M2Crypto.X509.X509 object.
    """
    if isinstance(string, str):
        string = string.encode("utf-8")
    bio = BIO.MemoryBuffer(string)
    return load_cert_bio(bio, format)


def load_cert_der_string(cert_str: Union[str, bytes]) -> X509:
    """
    Load certificate from a cert_str.

    :param cert_str: String containing a certificate in DER format.

    :return: M2Crypto.X509.X509 object.
    """
    if isinstance(cert_str, str):
        cert_str = cert_str.encode("utf-8")
    return load_cert_string(cert_str, FORMAT_DER)


class X509_Stack(object):
    """
    X509 Stack

    :warning: Do not modify the underlying OpenSSL stack
              except through this interface, or use any OpenSSL
              functions that do so indirectly. Doing so will get the
              OpenSSL stack and the internal pystack of this class out
              of sync, leading to python memory leaks, exceptions or
              even python crashes!
    """

    def __init__(
        self,
        stack: Optional[C.STACK_OF_X509] = None,
        _pyfree: int = 0,
        _pyfree_x509: int = 0,
    ) -> None:
        """
        :param stack: OpenSSL STACK_OF(X509)
        """
        self.pystack: List[X509] = []
        if stack is not None:
            self.stack = stack
            self._pyfree = _pyfree
            num = m2.sk_x509_num(self.stack)
            for i in range(num):
                self.pystack.append(
                    X509(m2.sk_x509_value(self.stack, i), _pyfree=_pyfree_x509)
                )
        else:
            self.stack = m2.sk_x509_new_null()
            self._pyfree = 1

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.sk_x509_free(self.stack)

    def __len__(self) -> int:
        return len(self.pystack)

    def __getitem__(self, idx: int) -> X509:
        return self.pystack[idx]

    def __iter__(self) -> Iterator[X509]:
        return iter(self.pystack)

    def _ptr(self) -> C.STACK_OF_X509:
        return self.stack

    def push(self, x509: X509) -> int:
        """
        push an X509 certificate onto the stack.

        :param x509: X509 object.

        :return: The number of X509 objects currently on the stack.
        """
        self.pystack.append(x509)
        return m2.sk_x509_push(self.stack, x509._ptr())

    def pop(self) -> Optional[X509]:
        """
        pop a certificate from the stack.

        :return: X509 object that was popped, or None if there is
                 nothing to pop.
        """
        x509_ptr = m2.sk_x509_pop(self.stack)
        if x509_ptr is None:
            # Sanity check: our Python list should also be empty.
            assert not self.pystack
            return None
        return self.pystack.pop()

    def as_der(self) -> bytes:
        """
        Return the stack as a DER encoded string
        """
        return m2.get_der_encoding_stack(self.stack)


class X509_Store_Context(object):
    """
    X509 Store Context
    """

    def __init__(self, x509_store_ctx: C.X509_STORE_CTX, _pyfree: int = 0) -> None:
        """

        :param x509_store_ctx: binary data for
              OpenSSL X509_STORE_CTX type
        """
        self.ctx = x509_store_ctx
        self._pyfree = _pyfree

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.x509_store_ctx_free(self.ctx)

    def _ptr(self) -> C.X509_STORE_CTX:
        return self.ctx

    def get_current_cert(self) -> Optional[X509]:
        """
        Get current X.509 certificate.

        :warning: The returned certificate is NOT refcounted, so you can not
                  rely on it being valid once the store context goes
                  away or is modified.
        """
        cert_ptr = m2.x509_store_ctx_get_current_cert(self.ctx)
        if cert_ptr is None:
            return None
        return X509(cert_ptr, _pyfree=0)

    def get_error(self) -> int:
        """
        Get error code.
        """
        return m2.x509_store_ctx_get_error(self.ctx)

    def get_error_depth(self) -> int:
        """
        Get error depth.
        """
        return m2.x509_store_ctx_get_error_depth(self.ctx)

    def get1_chain(self) -> X509_Stack:
        """
        Get certificate chain.

        :return: Reference counted (i.e. safe to use even after the store
                 context goes away) stack of certificates in the
                 chain as X509_Stack.
        """
        return X509_Stack(m2.x509_store_ctx_get1_chain(self.ctx), 1, 1)


def x509_store_default_cb(ok: int, ctx: C.X509_STORE_CTX) -> int:
    return ok


class X509_Store(object):
    """
    X509 Store
    """

    def __init__(self, store: Optional[C.X509_STORE] = None, _pyfree: int = 0) -> None:
        """
        :param store: binary data for OpenSSL X509_STORE_CTX type.
        """
        if store is not None:
            self.store = store
            self._pyfree = _pyfree
        else:
            self.store = m2.x509_store_new()
            self._pyfree = 1

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.x509_store_free(self.store)

    def _ptr(self) -> C.X509_STORE:
        return self.store

    def load_info(self, file: str) -> int:
        """
        :param file: filename

        :return: 1 on success, 0 on failure
        """
        return m2.x509_store_load_locations(self.store, file)

    load_locations = load_info

    def add_x509(self, x509: X509) -> int:
        """
        Add X509 certificate to the store.
        """
        return m2.x509_store_add_cert(self.store, x509._ptr())

    add_cert = add_x509

    def set_verify_cb(
        self, callback: Optional[Callable[[int, C.X509_STORE_CTX], int]] = None
    ) -> None:
        """
        Set callback which will be called when the store is verified.
        Wrapper over OpenSSL X509_STORE_set_verify_cb().

        :param callback:    Callable to specify verification options.
                            Type of the callable must be:
                            (int, X509_Store_Context) -> int.
                            If None: set the standard options.

        :note: compile-time or run-time errors in the callback would result
               in mysterious errors during verification, which could be hard
               to trace.

        :note: Python exceptions raised in callbacks do not propagate to
               verify() call.

        :return: None
        """
        if callback is None:
            callback = x509_store_default_cb

        if not callable(callback):
            raise X509Error("set_verify(): callback is not callable")

        m2.x509_store_set_verify_cb(self.store, callback)

    def set_flags(self, flags: int) -> int:
        """
        Set the verification flags for the X509Store
        Wrapper over OpenSSL X509_STORE_set_flags()

        :param flags: `VERIFICATION FLAGS` section of the
                      X509_VERIFY_PARAM_set_flags man page has
                      a complete description of values the flags
                      parameter can take.
                      Their M2Crypto equivalent is transformed following
                      the pattern: "X509_V_FLAG_XYZ" -> lowercase("VERIFY_XYZ")
        """
        return m2.x509_store_set_flags(self.store, flags)


def new_stack_from_der(der_string: bytes) -> X509_Stack:
    """
    Create a new X509_Stack from DER string.
    """
    stack_ptr = m2.make_stack_from_der_sequence(der_string)
    if stack_ptr is None:
        raise X509Error("Failed to create stack from DER sequence")
    return X509_Stack(stack_ptr, 1, 1)


class Request(object):
    """
    X509 Certificate Request.
    """

    def __init__(self, req: Optional[C.X509_REQ] = None, _pyfree: int = 0) -> None:
        if req is not None:
            self.req = req
            self._pyfree = _pyfree
        else:
            self.req = m2.x509_req_new()
            m2.x509_req_set_version(self.req, 0)
            self._pyfree = 1

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.x509_req_free(self.req)

    def _ptr(self) -> C.X509_REQ:
        return self.req

    def as_text(self) -> str:
        buf = BIO.MemoryBuffer()
        m2.x509_req_print(buf.bio_ptr(), self.req)
        return (buf.read_all() or b"").decode()

    def as_pem(self) -> bytes:
        buf = BIO.MemoryBuffer()
        m2.x509_req_write_pem(buf.bio_ptr(), self.req)
        return buf.read() or b""

    def as_der(self) -> bytes:
        buf = BIO.MemoryBuffer()
        m2.i2d_x509_req_bio(buf.bio_ptr(), self.req)
        return buf.read() or b""

    def save_pem(self, filename: str) -> int:
        with BIO.openfile(filename, "wb") as bio:
            return m2.x509_req_write_pem(bio.bio_ptr(), self.req)

    def save(self, filename: str, format: int = FORMAT_PEM) -> int:
        """
        Saves X.509 certificate request to a file. Default output
        format is PEM.

        :param filename: Name of the file the request will be s: aved to.

                       request. Either FORMAT_PEM or FORMAT_DER to save
                       in PEM or DER format. Raises V:  a lueError if an
                       unknown format is used.

        :return: 1 for success, 0 for failure.
                 The error code can be obtained by ERR_get_error.
        """
        with BIO.openfile(filename, "wb") as bio:
            if format == FORMAT_PEM:
                return m2.x509_req_write_pem(bio.bio_ptr(), self.req)
            elif format == FORMAT_DER:
                return m2.i2d_x509_req_bio(bio.bio_ptr(), self.req)
            else:
                raise ValueError(
                    "Unknown filetype. Must be either FORMAT_DER or FORMAT_PEM"
                )

    def get_pubkey(self) -> EVP.PKey:
        """
        Get the public key for the request.

        :return:     Public key from the request.
        """
        pkey_ptr = m2.x509_req_get_pubkey(self.req)
        if pkey_ptr is None:
            raise X509Error("Could not get public key from request")
        return EVP.PKey(pkey_ptr, _pyfree=1)

    def set_pubkey(self, pkey: EVP.PKey) -> int:
        """
        Set the public key for the request.

        :param pkey: Public key

        :return:     Return 1 for success and 0 for failure.
        """
        return m2.x509_req_set_pubkey(self.req, pkey.pkey)

    def get_subject(self) -> X509_Name:
        return X509_Name(m2.x509_req_get_subject_name(self.req), _pyfree=0)

    def set_subject_name(self, name: X509_Name) -> int:
        """
        Set subject name.

        :param name:    subjectName field.
        :return:    1 for success and 0 for failure
        """
        return m2.x509_req_set_subject_name(self.req, name._ptr())

    set_subject = set_subject_name

    def add_extensions(self, ext_stack: X509_Extension_Stack) -> int:
        """
        Add X509 extensions to this request.

        :param ext_stack: Stack of extensions to add.
        :return: 1 for success and 0 for failure
        """
        return m2.x509_req_add_extensions(self.req, ext_stack._ptr())

    def verify(self, pkey: EVP.PKey) -> int:
        """

        :param pkey: PKey to be verified
        :return: 1 for success and 0 for failure
        """
        return m2.x509_req_verify(self.req, pkey.pkey)

    def sign(self, pkey: EVP.PKey, md: str) -> int:
        """

        :param pkey: PKey to be signed
        :param md: used algorigthm
        :return: 1 for success and 0 for failure
        """
        mda = getattr(m2, md, None)
        if mda is None:
            raise ValueError("unknown message digest", md)
        return m2.x509_req_sign(self.req, pkey.pkey, mda())

    def get_version(self) -> int:
        """
        Get the version of the request.
        :return: Version number.
        """
        return m2.x509_req_get_version(self.req)

    def set_version(self, ver: int) -> int:
        """
        Set the version of the request.
        :param ver: Version number.
        :return: 1 for success, 0 for failure.
        """
        return m2.x509_req_set_version(self.req, ver)


def load_request(file: Union[str, bytes], format: int = FORMAT_PEM) -> Request:
    """
    Load certificate request from file.

    :param file: Name of file containing certificate request in
                 either PEM or DER format.
    :param format: Describes the format of the file to be loaded,
                   either PEM or DER. (using constants FORMAT_PEM
                   and FORMAT_DER)
    :return: Request object.
    """
    if isinstance(file, bytes):
        file = file.decode()
    with BIO.openfile(file) as f:
        if format == FORMAT_PEM:
            cptr = m2.x509_req_read_pem(f.bio_ptr())
        elif format == FORMAT_DER:
            cptr = m2.d2i_x509_req(f.bio_ptr())
        else:
            raise ValueError(
                "Unknown filetype. Must be either FORMAT_PEM or FORMAT_DER"
            )

    if cptr is None:
        raise X509Error("Failed to load request from file")

    return Request(cptr, 1)


def load_request_bio(bio: BIO.BIO, format: int = FORMAT_PEM) -> Request:
    """
    Load certificate request from a bio.

    :param bio: BIO pointing at a certificate request in
                either DER or PEM format.
    :param format: Describes the format of the request to be loaded,
                   either PEM or DER. (using constants FORMAT_PEM
                   and FORMAT_DER)
    :return: M2Crypto.X509.Request object.
    """
    if format == FORMAT_PEM:
        cptr = m2.x509_req_read_pem(bio._ptr())
    elif format == FORMAT_DER:
        cptr = m2.d2i_x509_req(bio._ptr())
    else:
        raise ValueError("Unknown format. Must be either FORMAT_DER or FORMAT_PEM")
    if cptr is None:
        raise X509Error("Failed to load request from BIO")
    return Request(cptr, _pyfree=1)


def load_request_string(string: Union[str, bytes], format: int = FORMAT_PEM) -> Request:
    """
    Load certificate request from a cert_str.

    :param cert_str: String containing a certificate request in
                     either DER or PEM format.
    :param format: Describes the format of the request to be loaded,
                   either PEM or DER. (using constants FORMAT_PEM
                   and FORMAT_DER)

    :return: M2Crypto.X509.Request object.
    """
    if isinstance(string, str):
        string = string.encode("utf-8")
    bio = BIO.MemoryBuffer(string)
    return load_request_bio(bio, format)


def load_request_der_string(cert_str: Union[str, bytes]) -> Request:
    """
    Load certificate request from a cert_str.

    :param cert_str: String containing a certificate request in DER format.
    :return: M2Crypto.X509.Request object.
    """
    cert_str = cert_str.encode() if isinstance(cert_str, str) else cert_str
    return load_request_string(cert_str, FORMAT_DER)


class CRL(object):
    """
    X509 Certificate Revocation List
    """

    def __init__(self, crl: Optional[C.X509_CRL] = None, _pyfree: int = 0) -> None:
        """

        :param crl: binary representation of
               the underlying OpenSSL X509_CRL object.
        """
        if crl is not None:
            self.crl = crl
            self._pyfree = _pyfree
        else:
            self.crl = m2.x509_crl_new()
            self._pyfree = 1

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            m2.x509_crl_free(self.crl)

    def as_text(self) -> str:
        """
        Return CRL in PEM format in a string.

        :return: String containing the CRL in PEM format.
        """
        buf = BIO.MemoryBuffer()
        m2.x509_crl_print(buf.bio_ptr(), self.crl)
        return (buf.read_all() or b"").decode()


def load_crl(file: str) -> CRL:
    """
    Load CRL from file.

    :param file: Name of file containing CRL in PEM format.

    :return: M2Crypto.X509.CRL object.
    """
    with BIO.openfile(file) as f:
        cptr = m2.x509_crl_read_pem(f.bio_ptr())
    if cptr is None:
        raise X509Error("Failed to load CRL from file")
    return CRL(cptr, 1)



"""M2Crypto wrapper for OpenSSL BIO API.

Copyright (c) 1999-2004 Ng Pheng Siong. All rights reserved."""

import io
import logging
from typing import Callable, IO, Iterable, Optional, Union, TYPE_CHECKING

from M2Crypto import m2, types as C

if TYPE_CHECKING:
    import M2Crypto.SSL as SSL

log = logging.getLogger("BIO")


class BIOError(ValueError):
    pass


m2.bio_init(BIOError)


class BIO(object):
    """Abstract object interface to the BIO API."""

    def __init__(
        self,
        bio: Optional[C.BIO] = None,
        _pyfree: int = 0,
        _close_cb: Optional[Callable] = None,
    ) -> None:
        self.bio = bio
        self._pyfree = _pyfree
        self._close_cb = _close_cb
        self.closed = 0
        self.write_closed = 0

    def __del__(self):
        if self._pyfree and self.bio:
            m2.bio_free(self.bio)

    def _ptr(self):
        return self.bio

    # Deprecated.
    bio_ptr = _ptr

    def fileno(self) -> int:
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        return m2.bio_get_fd(self.bio)

    def readable(self) -> bool:
        return not self.closed

    def read(self, size: Optional[int] = None) -> Union[bytes, bytearray]:
        if not self.readable():
            raise IOError("cannot read")
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        if size is None:
            buf = bytearray()
            while 1:
                data = m2.bio_read(self.bio, 4096)
                if not data:
                    break
                buf += data
            return buf
        elif size == 0:
            return b""
        elif size < 0:
            raise ValueError("read count is negative")
        else:
            return bytes(m2.bio_read(self.bio, size))

    def readline(self, size: int = 4096) -> bytes:
        if not self.readable():
            raise IOError("cannot read")
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        buf = m2.bio_gets(self.bio, size)
        return b"" if buf is None else buf

    def readlines(
        self, sizehint: Union[str, bytes, int] = "ignored"
    ) -> Iterable[bytes]:
        if not self.readable():
            raise IOError("cannot read")
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        lines = []
        while 1:
            buf = m2.bio_gets(self.bio, 4096)
            if buf is None:
                break
            lines.append(buf)
        return lines

    def writeable(self) -> bool:
        return (not self.closed) and (not self.write_closed)

    def write(self, data: Union[str, bytes]) -> int:
        """Write data to BIO.

        :return: either data written, or [0, -1] for nothing written,
                 -2 not implemented
        """
        if not self.writeable():
            raise IOError("cannot write")
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        if isinstance(data, str):
            data = data.encode("utf8")
        return m2.bio_write(self.bio, data)

    def write_close(self) -> None:
        self.write_closed = 1

    def flush(self) -> None:
        """Flush the buffers.

        :return: 1 for success, and 0 or -1 for failure
        """
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        m2.bio_flush(self.bio)

    def reset(self) -> int:
        """Set the bio to its initial state.

        :return: 1 for success, and 0 or -1 for failure
        """
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        return m2.bio_reset(self.bio)

    def close(self) -> None:
        self.closed = 1
        if self._close_cb:
            self._close_cb()

    def should_retry(self) -> int:
        """
        Can the call be attempted again, or was there an error
        ie do_handshake

        """
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        return m2.bio_should_retry(self.bio)

    def should_read(self) -> int:
        """Should we read more data?"""
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        return m2.bio_should_read(self.bio)

    def should_write(self) -> int:
        """Should we write more data?"""
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        return m2.bio_should_write(self.bio)

    def tell(self):
        """Return the current offset."""
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        return m2.bio_tell(self.bio)

    def seek(self, off):
        """Seek to the specified absolute offset."""
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        return m2.bio_seek(self.bio, off)

    def __enter__(self):
        return self

    def __exit__(self, *args) -> None:
        self.close()


class MemoryBuffer(BIO):
    """Object interface to BIO_s_mem.

    Empirical testing suggests that this class performs less well than
    cStringIO, because cStringIO is implemented in C, whereas this class
    is implemented in Python. Thus, the recommended practice is to use
    cStringIO for regular work and convert said cStringIO object to
    a MemoryBuffer object only when necessary.
    """

    def __init__(self, data: Optional[bytes] = None) -> None:
        super().__init__()
        if data is not None and not isinstance(data, bytes):
            raise TypeError(
                "data must be bytes or None, not %s" % (type(data).__name__,)
            )
        self.bio = m2.bio_new(m2.bio_s_mem())
        self._pyfree = 1
        if data is not None:
            m2.bio_write(self.bio, data)

    def __len__(self) -> int:
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        return m2.bio_ctrl_pending(self.bio)

    def read(self, size: Optional[int] = None) -> bytes:
        m2.err_clear_error()
        if not self.readable():
            raise IOError("cannot read")
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        if size is None:
            return m2.bio_read(self.bio, m2.bio_ctrl_pending(self.bio))
        if size < 0:
            raise ValueError("read count is negative")
        return m2.bio_read(self.bio, size)

    # Backwards-compatibility.
    getvalue = read_all = read

    def write_close(self) -> None:
        super().write_close()
        if self.bio is None:
            raise BIOError("Uninitialized self.bio")
        m2.bio_set_mem_eof_return(self.bio, 0)

    close = write_close


class File(BIO):
    """Object interface to BIO_s_pyfd.

    This class interfaces Python to OpenSSL functions that expect BIO. For
    general file manipulation in Python, use Python's builtin file object.
    """

    def __init__(
        self,
        pyfile_arg: Union[IO, str],
        close_pyfile: int = 1,
        mode: str = "rb",
    ) -> None:
        super().__init__(_pyfree=1)

        pyfile: IO
        if isinstance(pyfile_arg, str):
            pyfile = open(pyfile_arg, mode)
        else:
            pyfile = pyfile_arg

        pyfile.flush()
        self.fname = getattr(pyfile, "name", "<unnamed>")
        self.pyfile = pyfile
        # Be wary of https://github.com/openssl/openssl/pull/1925
        # BIO_new_fd is NEVER to be used before OpenSSL 1.1.1
        if hasattr(m2, "bio_new_pyfd"):
            self.bio = m2.bio_new_pyfd(pyfile.fileno(), m2.bio_noclose)
        else:
            self.bio = m2.bio_new_pyfile(pyfile, m2.bio_noclose)

        self.close_pyfile = close_pyfile
        self.closed = False

    def flush(self) -> None:
        super().flush()
        self.pyfile.flush()

    def close(self) -> None:
        self.flush()
        super().close()
        if self.close_pyfile:
            self.pyfile.close()

    def reset(self) -> int:
        """Set the bio to its initial state.

        :return: 0 for success, and -1 for failure
        """
        return super().reset()


def openfile(filename: Union[str, bytes], mode: str = "rb") -> File:
    try:
        f = open(filename, mode)
    except IOError as ex:
        raise BIOError(ex.args)

    return File(f)


class IOBuffer(BIO):
    """Object interface to BIO_f_buffer.

    Its principal function is to be BIO_push()'ed on top of a BIO_f_ssl, so
    that makefile() of said underlying SSL socket works.
    """

    def __init__(self, under_bio: BIO, mode: str = "rwb", _pyfree: int = 1) -> None:
        super().__init__(_pyfree=_pyfree)
        self.io = m2.bio_new(m2.bio_f_buffer())
        self.bio = m2.bio_push(self.io, under_bio._ptr())
        # This reference keeps the underlying BIO alive while we're not closed.
        self._under_bio = under_bio
        if "w" in mode:
            self.write_closed = 0
        else:
            self.write_closed = 1

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            if self.bio:
                m2.bio_pop(self.bio)
        if self.io:
            m2.bio_free(self.io)

    def close(self) -> None:
        BIO.close(self)


class CipherStream(BIO):
    """Object interface to BIO_f_cipher."""

    SALT_LEN = m2.PKCS5_SALT_LEN

    def __init__(self, obio: BIO) -> None:
        super().__init__(_pyfree=1)
        self.obio = obio
        self.bio = m2.bio_new(m2.bio_f_cipher())
        self.closed = 0

    def __del__(self) -> None:
        if not getattr(self, "closed", 1):
            self.close()

    def close(self) -> None:
        if self.bio:
            m2.bio_pop(self.bio)
            m2.bio_free(self.bio)
        self.closed = 1

    def write_close(self) -> None:
        self.obio.write_close()

    def set_cipher(
        self,
        algo: str,
        key: Union[str, bytes],
        iv: Union[str, bytes],
        op: int,
    ) -> None:
        cipher = getattr(m2, algo, None)
        if cipher is None:
            raise ValueError("unknown cipher", algo)
        else:
            if not isinstance(key, bytes):
                key = key.encode("utf8")
            if not isinstance(iv, bytes):
                iv = iv.encode("utf8")
        try:
            assert self.bio is not None
            m2.bio_set_cipher(self.bio, cipher(), key, iv, int(op))
        except (OSError, ValueError) as ex:
            raise BIOError("BIOError: {}".format(str(ex)))
        assert self.bio is not None
        m2.bio_push(self.bio, self.obio._ptr())


class SSLBio(BIO):
    """Object interface to BIO_f_ssl."""

    def __init__(self, _pyfree: int = 1) -> None:
        super().__init__(_pyfree=_pyfree)
        self.bio = m2.bio_new(m2.bio_f_ssl())
        self.closed = 0

    def set_ssl(self, conn: "SSL.Connection", close_flag: int = m2.bio_noclose) -> None:
        """
        Sets the bio to the SSL pointer which is
        contained in the connection object.
        """
        self._pyfree = 0
        assert self.bio is not None
        m2.bio_set_ssl(self.bio, conn.ssl, close_flag)  # type: ignore[arg-type]
        if close_flag == m2.bio_noclose:
            conn.set_ssl_close_flag(m2.bio_close)

    def do_handshake(self) -> int:
        """Do the handshake.

        Return 1 if the handshake completes
        Return 0 or a negative number if there is a problem
        """
        assert self.bio is not None
        return m2.bio_do_handshake(self.bio)



"""SSL callbacks

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

import sys

from M2Crypto import X509, m2, types as C
from typing import List

__all__ = [
    "unknown_issuer",
    "ssl_verify_callback_stub",
    "ssl_verify_callback",
    "ssl_verify_callback_allow_unknown_ca",
    "ssl_info_callback",
]


def ssl_verify_callback_stub(ssl_ctx_ptr, x509_ptr, errnum, errdepth, ok):
    # Deprecated
    return ok


unknown_issuer: List[int] = [
    m2.X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT,
    m2.X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY,
    m2.X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE,
    m2.X509_V_ERR_CERT_UNTRUSTED,
]


def ssl_verify_callback(
    ssl_ctx_ptr: C.SSL_CTX,
    x509_ptr: C.X509,
    errnum: int,
    errdepth: int,
    ok: int,
) -> int:
    from M2Crypto.SSL.Context import Context, ctxmap

    ssl_ctx = ctxmap()[id(ssl_ctx_ptr)]
    if errnum in unknown_issuer:
        if ssl_ctx.get_allow_unknown_ca():
            sys.stderr.write(
                "policy: %s: permitted...\n" % (m2.x509_get_verify_error(errnum))
            )
            sys.stderr.flush()
            ok = 1
    # CRL checking goes here...
    if ok:
        if ssl_ctx.get_verify_depth() >= errdepth:
            ok = 1
        else:
            ok = 0
    return ok


def ssl_verify_callback_allow_unknown_ca(
    ok: int, store: X509.X509_Store_Context
) -> int:
    """
    Callback that allows unknown CA errors.
    This version relies on a corrected SWIG typemap to receive a valid 'store' object.
    """
    store_ptr = store.ctx
    errnum = m2.x509_store_ctx_get_error(store_ptr)

    if errnum in unknown_issuer:
        # It's an error we want to ignore. Clear the error from the
        # context and return 1 to override the failure.
        m2.x509_store_ctx_set_error(store_ptr, m2.X509_V_OK)
        return 1

    # For any other error, respect the original verification status.
    return ok


# Cribbed from OpenSSL's apps/s_cb.c.
def ssl_info_callback(where: int, ret: int, ssl_ptr: C.SSL) -> None:
    where_int = where & ~m2.SSL_ST_MASK
    if where_int & m2.SSL_ST_CONNECT:
        state = "SSL connect"
    elif where_int & m2.SSL_ST_ACCEPT:
        state = "SSL accept"
    else:
        state = "SSL state unknown"

    if where & m2.SSL_CB_LOOP:
        sys.stderr.write("LOOP: %s: %s\n" % (state, m2.ssl_get_state_v(ssl_ptr)))
        sys.stderr.flush()
        return

    if where & m2.SSL_CB_EXIT:
        if not ret:
            sys.stderr.write("FAILED: %s: %s\n" % (state, m2.ssl_get_state_v(ssl_ptr)))
            sys.stderr.flush()
        else:
            sys.stderr.write("INFO: %s: %s\n" % (state, m2.ssl_get_state_v(ssl_ptr)))
            sys.stderr.flush()
        return

    if where & m2.SSL_CB_ALERT:
        # Use a new variable for the alert operation string
        alert_op = "read" if (where & m2.SSL_CB_READ) else "write"
        sys.stderr.write(
            "ALERT: %s: %s: %s\n"
            % (
                alert_op,
                m2.ssl_get_alert_type_v(ret),
                m2.ssl_get_alert_desc_v(ret),
            )
        )
        sys.stderr.flush()
        return



"""SSL Exceptions

Copyright (c) 1999-2004 Ng Pheng Siong. All rights reserved."""

import socket

__all__ = ["SSLError", "SSLTimeoutError"]


class SSLError(Exception):
    pass


class SSLTimeoutError(SSLError, socket.timeout):
    pass



"""M2Crypto SSL services.

Copyright (c) 1999-2004 Ng Pheng Siong. All rights reserved."""

import socket, os

# M2Crypto
from M2Crypto import _m2crypto as m2

from .SSLError import SSLError, SSLTimeoutError

m2.ssl_init(SSLError, SSLTimeoutError)

# M2Crypto.SSL
from .Cipher import Cipher, Cipher_Stack
from .Connection import Connection
from .Context import Context
from .SSLServer import SSLServer, ThreadingSSLServer

if os.name != "nt":
    from .SSLServer import ForkingSSLServer
from .timeout import (
    timeout,
    struct_to_timeout,
    struct_size,
)

verify_none: int = m2.SSL_VERIFY_NONE
verify_peer: int = m2.SSL_VERIFY_PEER
verify_fail_if_no_peer_cert: int = m2.SSL_VERIFY_FAIL_IF_NO_PEER_CERT
verify_client_once: int = m2.SSL_VERIFY_CLIENT_ONCE
verify_crl_check_chain: int = m2.VERIFY_CRL_CHECK_CHAIN
verify_crl_check_leaf: int = m2.VERIFY_CRL_CHECK_LEAF

SSL_SENT_SHUTDOWN: int = m2.SSL_SENT_SHUTDOWN
SSL_RECEIVED_SHUTDOWN: int = m2.SSL_RECEIVED_SHUTDOWN

op_all: int = m2.SSL_OP_ALL
op_no_sslv2: int = m2.SSL_OP_NO_SSLv2



"""
M2Crypto utility routines.

NOTHING IN THIS MODULE IS GUARANTEED TO BE STABLE, USED ONLY FOR
INTERNAL PURPOSES OF M2CRYPTO.

Copyright (c) 1999-2004 Ng Pheng Siong. All rights reserved.

Portions created by Open Source Applications Foundation (OSAF) are
Copyright (C) 2004 OSAF. All Rights Reserved.
"""

import binascii
import logging
import platform
import struct
import sys
import struct
import unittest

from M2Crypto import m2
from typing import (
    Callable,
    Optional,
    TextIO,
    Tuple,
    Union,
)

# see https://github.com/python/typeshed/issues/222
AddrType = Union[Tuple[str, int], str]
# COMPATIBILITY: Literal is only Python 3.8+
# ModeStr = Literal['r', 'w', 'rw', 'rb', 'wb', 'rwb']
ModeStr = str

log = logging.getLogger("util")


class UtilError(Exception):
    pass


m2.util_init(UtilError)


def is_32bit() -> bool:
    # or alternatively (slightly slower)
    # (struct.calcsize("P") * 8) == 32
    return not (sys.maxsize > 2**32)


def expectedFailureIf(condition: bool) -> Callable:
    """The test is marked as an expectedFailure if the condition is satisfied."""

    def wrapper(func):
        if condition:
            return unittest.expectedFailure(func)
        else:
            return func

    return wrapper


def pkcs5_pad(data: str, blklen: int = 8) -> str:
    pad = 8 - (len(data) % 8)
    return data + chr(pad) * pad


def pkcs7_pad(data: str, blklen: int) -> str:
    if blklen > 255:
        raise ValueError("illegal block size")
    pad = blklen - (len(data) % blklen)
    return data + chr(pad) * pad


def bin_to_hex(b: bytes) -> str:
    return binascii.b2a_base64(b)[:-1].decode()


def octx_to_num(x: bytes) -> int:
    return int(binascii.hexlify(x), 16)


def genparam_callback(p: int, n: int) -> int:
    ch = [".", "+", "*", "\n"]
    sys.stdout.write(ch[p])
    sys.stdout.flush()
    return 1


def quiet_genparam_callback(p: int, n: int) -> int:
    return 1


def passphrase_callback(
    v: bool,
    prompt1: str = "Enter passphrase:",
    prompt2: str = "Verify passphrase:",
) -> Optional[str]:
    from getpass import getpass

    while 1:
        try:
            p1 = getpass(prompt1)
            if v:
                p2 = getpass(prompt2)
                if p1 == p2:
                    break
            else:
                break
        except KeyboardInterrupt:
            return None
    return p1


def no_passphrase_callback(*args) -> str:
    return ""



"""
M2Crypto threading support, required for multithreaded applications.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

# M2Crypto
from M2Crypto import m2


def init() -> None:
    """
    Initialize threading support.
    """
    m2.threading_init()


def cleanup() -> None:
    """
    End and cleanup threading support.
    """
    m2.threading_cleanup()



"""
M2Crypto wrapper for OpenSSL DSA API.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved.

Portions created by Open Source Applications Foundation (OSAF) are
Copyright (C) 2004 OSAF. All Rights Reserved.
"""

from M2Crypto import BIO, m2, util, types as C
from typing import Callable, NoReturn, Tuple, Union  # noqa


class DSAError(Exception):
    pass


m2.dsa_init(DSAError)


class DSA:
    """
    This class is a context supporting DSA key and parameter
    values, signing and verifying.

    Simple example::

        from M2Crypto import EVP, DSA, util

        message = 'Kilroy was here!'
        md = EVP.MessageDigest('sha256')
        md.update(message)
        digest = md.final()

        dsa = DSA.gen_params(1024, cb)
        dsa.gen_key()
        r, s = dsa.sign(digest)
        good = dsa.verify(digest, r, s)
        if good:
            print('  ** success **')
        else:
            print('  ** verification failed **')
    """

    def __init__(self, dsa: C.DSA, _pyfree: int = 0) -> None:
        """
        Use one of the factory functions to create an instance.
        :param dsa: binary representation of OpenSSL DSA type
        """
        assert m2.dsa_type_check(dsa), "'dsa' type error"
        self.dsa = dsa
        self._pyfree = _pyfree

    @staticmethod
    def m2_dsa_free(dsa: C.DSA) -> None:
        m2.dsa_free(dsa)

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            self.m2_dsa_free(self.dsa)

    def __len__(self) -> int:
        """
        Return the key length.

        :return:  the DSA key length in bits
        """
        assert m2.dsa_type_check(self.dsa), "'dsa' type error"
        return m2.dsa_keylen(self.dsa)

    def __getattr__(self, name: str) -> bytes:
        """
        Return specified DSA parameters and key values.

        :param name: name of variable to be returned.  Must be
                     one of 'p', 'q', 'g', 'pub', 'priv'.
        :return:     value of specified variable (a "byte string")
        """
        if name in ["p", "q", "g", "pub", "priv"]:
            method = getattr(m2, "dsa_get_%s" % (name,))
            assert m2.dsa_type_check(self.dsa), "'dsa' type error"
            return method(self.dsa)
        else:
            raise AttributeError

    def __setattr__(self, name: str, value: bytes) -> None:
        if name in ["p", "q", "g"]:
            raise DSAError("set (p, q, g) via set_params()")
        elif name in ["pub", "priv"]:
            raise DSAError("generate (pub, priv) via gen_key()")
        else:
            self.__dict__[name] = value

    def set_params(self, p: bytes, q: bytes, g: bytes) -> None:
        """
        Set new parameters.

        :param p: MPI binary representation ... format that consists of
                  the number's length in bytes represented as a 4-byte
                  big-endian number, and the number itself in big-endian
                  format, where the most significant bit signals
                  a negative number (the representation of numbers with
                  the MSB set is prefixed with null byte).
        :param q: ditto
        :param g: ditto

        @warning: This does not change the private key, so it may be
                  unsafe to use this method. It is better to use
                  gen_params function to create a new DSA object.
        """
        m2.dsa_set_pqg(self.dsa, p, q, g)

    def gen_key(self) -> None:
        """
        Generate a key pair.
        """
        assert m2.dsa_type_check(self.dsa), "'dsa' type error"
        m2.dsa_gen_key(self.dsa)

    def save_params(self, filename: Union[str, bytes]) -> int:
        """
        Save the DSA parameters to a file.

        :param filename: Save the DSA parameters to this file.
        :return:         1 (true) if successful
        """
        with BIO.openfile(filename, "wb") as bio:
            ret = m2.dsa_write_params_bio(self.dsa, bio._ptr())

        return ret

    def save_params_bio(self, bio: BIO.BIO) -> int:
        """
        Save DSA parameters to a BIO object.

        :param bio: Save DSA parameters to this object.
        :return:    1 (true) if successful
        """
        return m2.dsa_write_params_bio(self.dsa, bio._ptr())

    def save_key(
        self,
        filename: Union[str, bytes],
        cipher: str = "aes_128_cbc",
        callback: Callable = util.passphrase_callback,
    ) -> int:
        """
        Save the DSA key pair to a file.

        :param filename: Save the DSA key pair to this file.
        :param cipher:   name of symmetric key algorithm and mode
                         to encrypt the private key.
        :return:         1 (true) if successful
        """
        with BIO.openfile(filename, "wb") as bio:
            ret = self.save_key_bio(bio, cipher, callback)

        return ret

    def save_key_bio(
        self,
        bio: BIO.BIO,
        cipher: str = "aes_128_cbc",
        callback: Callable = util.passphrase_callback,
    ) -> int:
        """
        Save DSA key pair to a BIO object.

        :param bio:    Save DSA parameters to this object.
        :param cipher: name of symmetric key algorithm and mode
                       to encrypt the private key.
        :return:       1 (true) if successful
        """
        if cipher is None:
            return m2.dsa_write_key_bio_no_cipher(self.dsa, bio._ptr(), callback)
        else:
            ciph = getattr(m2, cipher, None)
            if ciph is None:
                raise DSAError("no such cipher: %s" % cipher)
            else:
                ciph = ciph()
            return m2.dsa_write_key_bio(self.dsa, bio._ptr(), ciph, callback)

    def save_pub_key(self, filename: Union[str, bytes]) -> int:
        """
        Save the DSA public key (with parameters) to a file.

        :param filename: Save DSA public key (with parameters)
                         to this file.
        :return:         1 (true) if successful
        """
        with BIO.openfile(filename, "wb") as bio:
            ret = self.save_pub_key_bio(bio)

        return ret

    def save_pub_key_bio(self, bio: BIO.BIO) -> int:
        """
        Save DSA public key (with parameters) to a BIO object.

        :param bio: Save DSA public key (with parameters)
                    to this object.
        :return:  1 (true) if successful
        """
        return m2.dsa_write_pub_key_bio(self.dsa, bio._ptr())

    def sign(self, digest: bytes) -> Tuple[bytes, bytes]:
        """
        Sign the digest.

        :param digest: SHA-1 hash of message (same as output
                       from MessageDigest, a "byte string")
        :return:       DSA signature, a tuple of two values, r and s,
                       both "byte strings".
        """
        assert self.check_key(), "key is not initialised"
        return m2.dsa_sign(self.dsa, digest)

    def verify(self, digest: bytes, r: bytes, s: bytes) -> int:
        """
        Verify a newly calculated digest against the signature
        values r and s.

        :param digest: SHA-1 hash of message (same as output
                       from MessageDigest, a "byte string")
        :param r:      r value of the signature, a "byte string"
        :param s:      s value of the signature, a "byte string"
        :return:       1 (true) if verify succeeded, 0 if failed
        """
        assert self.check_key(), "key is not initialised"
        return m2.dsa_verify(self.dsa, digest, r, s)

    def sign_asn1(self, digest):
        assert self.check_key(), "key is not initialised"
        return m2.dsa_sign_asn1(self.dsa, digest)

    def verify_asn1(self, digest, blob):
        assert self.check_key(), "key is not initialised"
        return m2.dsa_verify_asn1(self.dsa, digest, blob)

    def check_key(self):
        """
        Check to be sure the DSA object has a valid private key.

        :return:  1 (true) if a valid private key
        """
        assert m2.dsa_type_check(self.dsa), "'dsa' type error"
        return m2.dsa_check_key(self.dsa)


class DSA_pub(DSA):
    """
    This class is a DSA context that only supports a public key
    and verification.  It does NOT support a private key or
    signing.

    """

    def sign(self, digest: bytes = b"") -> NoReturn:
        raise DSAError("DSA_pub object has no private key")

    sign_asn1 = sign

    def check_key(self) -> int:
        """
        :return: does DSA_pub contain a pub key?
        """
        return m2.dsa_check_pub_key(self.dsa)

    save_key: Callable = DSA.save_pub_key

    save_key_bio: Callable = DSA.save_pub_key_bio


# --------------------------------------------------------------
# factories and other functions


def gen_params(
    bits: int,
    callback: Callable[[int, int], int] = util.genparam_callback,
) -> DSA:
    """
    Factory function that generates DSA parameters and
    instantiates a DSA object from the output.

    :param bits: The length of the prime to be generated. If
                 'bits' < 512, it is set to 512.
    :param callback: A Python callback object that will be
                 invoked during parameter generation; it usual
                 purpose is to provide visual feedback.
    :return:  instance of DSA.
    """
    dsa = m2.dsa_generate_parameters(bits, callback)
    return DSA(dsa, 1)


def set_params(p: bytes, q: bytes, g: bytes) -> DSA:
    """
    Factory function that instantiates a DSA object with DSA
    parameters.

    :param p: value of p, a "byte string"
    :param q: value of q, a "byte string"
    :param g: value of g, a "byte string"
    :return:  instance of DSA.
    """
    dsa = m2.dsa_new()
    m2.dsa_set_pqg(dsa, p, q, g)
    return DSA(dsa, 1)


def load_params(
    file: Union[str, bytes],
    callback: Callable = util.passphrase_callback,
) -> DSA:
    """
    Factory function that instantiates a DSA object with DSA
    parameters from a file.

    :param file:     Names the file (a path) that contains the PEM
                     representation of the DSA parameters.
    :param callback: A Python callback object that will be
                     invoked if the DSA parameters file is
                     passphrase-protected.
    :return:         instance of DSA.
    """
    with BIO.openfile(file) as bio:
        ret = load_params_bio(bio, callback)

    return ret


def load_params_bio(bio: BIO.BIO, callback: Callable = util.passphrase_callback) -> DSA:
    """
    Factory function that instantiates a DSA object with DSA
    parameters from a M2Crypto.BIO object.

    :param bio:      Contains the PEM representation of the DSA
                     parameters.
    :param callback: A Python callback object that will be
                     invoked if the DSA parameters file is
                     passphrase-protected.
    :return:         instance of DSA.
    """
    dsa = m2.dsa_read_params(bio._ptr(), callback)
    if dsa is None:
        raise DSAError("failed to load DSA parameters")
    return DSA(dsa, 1)


def load_key(
    file: Union[str, bytes],
    callback: Callable = util.passphrase_callback,
) -> DSA:
    """
    Factory function that instantiates a DSA object from a
    PEM encoded DSA key pair.

    :param file:     Names the file (a path) that contains the PEM
                     representation of the DSA key pair.
    :param callback: A Python callback object that will be
                     invoked if the DSA key pair is
                     passphrase-protected.
    :return:         instance of DSA.
    """
    with BIO.openfile(file) as bio:
        ret = load_key_bio(bio, callback)

    return ret


def load_key_bio(bio: BIO.BIO, callback: Callable = util.passphrase_callback) -> DSA:
    """
    Factory function that instantiates a DSA object from a
    PEM encoded DSA key pair.

    :param bio:      Contains the PEM representation of the DSA
                     key pair.
    :param callback: A Python callback object that will be
                     invoked if the DSA key pair is
                     passphrase-protected.
    :return:         instance of DSA.
    """
    dsa = m2.dsa_read_key(bio._ptr(), callback)
    if dsa is None:
        raise DSAError("failed to load DSA key")
    return DSA(dsa, 1)


def pub_key_from_params(p: bytes, q: bytes, g: bytes, pub: bytes) -> DSA_pub:
    """
    Factory function that instantiates a DSA_pub object using
    the parameters and public key specified.

    :param p: value of p
    :param q: value of q
    :param g: value of g
    :param pub: value of the public key
    :return:  instance of DSA_pub.
    """
    dsa = m2.dsa_new()
    m2.dsa_set_pqg(dsa, p, q, g)
    m2.dsa_set_pub(dsa, pub)
    return DSA_pub(dsa, 1)


def load_pub_key(
    file: Union[str, bytes],
    callback: Callable = util.passphrase_callback,
) -> DSA_pub:
    """
    Factory function that instantiates a DSA_pub object using
    a DSA public key contained in PEM file.  The PEM file
    must contain the parameters in addition to the public key.

    :param file:     Names the file (a path) that contains the PEM
                     representation of the DSA public key.
    :param callback: A Python callback object that will be
                     invoked should the DSA public key be
                     passphrase-protected.
    :return:         instance of DSA_pub.
    """
    with BIO.openfile(file) as bio:
        ret = load_pub_key_bio(bio, callback)

    return ret


def load_pub_key_bio(
    bio: BIO.BIO, callback: Callable = util.passphrase_callback
) -> DSA_pub:
    """
    Factory function that instantiates a DSA_pub object using
    a DSA public key contained in PEM format.  The PEM
    must contain the parameters in addition to the public key.

    :param bio:      Contains the PEM representation of the DSA
                     public key (with params).
    :param callback: A Python callback object that will be
                     invoked should the DSA public key be
                     passphrase-protected.
    :return:         instance of DSA_pub.
    """
    dsapub = m2.dsa_read_pub_key(bio._ptr(), callback)
    if dsapub is None:
        raise DSAError("failed to load DSA public key")
    return DSA_pub(dsapub, 1)



"""SSL Session

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

__all__ = ["Session", "load_session"]

from M2Crypto import BIO, Err, m2, types as C
from typing import Union, cast, TYPE_CHECKING  # noqa


class Session(object):

    def __init__(self, session, _pyfree: int = 0) -> None:
        assert session is not None
        self.session = session
        self._pyfree = _pyfree

    @staticmethod
    def m2_ssl_session_free(sess: C.SSL_SESSION) -> None:
        m2.ssl_session_free(sess)

    def __del__(self) -> None:
        if getattr(self, "_pyfree", 0):
            self.m2_ssl_session_free(self.session)

    def _ptr(self) -> bytes:
        return self.session

    def as_text(self) -> bytes:
        buf = BIO.MemoryBuffer()
        m2.ssl_session_print(buf.bio_ptr(), self.session)
        return buf.read_all()

    def as_der(self) -> bytes:
        buf = BIO.MemoryBuffer()
        m2.i2d_ssl_session(buf.bio_ptr(), self.session)
        return buf.read_all()

    def write_bio(self, bio: BIO.BIO) -> int:
        return m2.ssl_session_write_pem(bio.bio_ptr(), self.session)

    def get_time(self) -> int:
        return m2.ssl_session_get_time(self.session)

    def set_time(self, t: int) -> int:
        return m2.ssl_session_set_time(self.session, t)

    def get_timeout(self) -> int:
        return m2.ssl_session_get_timeout(self.session)

    def set_timeout(self, t: int) -> int:
        return m2.ssl_session_set_timeout(self.session, t)


def load_session(pemfile: Union[str, bytes]) -> Session:
    with BIO.openfile(pemfile) as f:
        cptr = m2.ssl_session_read_pem(f.bio_ptr())

    return Session(cptr, 1)



"""M2Crypto wrapper for OpenSSL Error API.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

from M2Crypto import BIO, m2, util  # noqa
from typing import Optional  # noqa


def get_error() -> Optional[str]:
    err = BIO.MemoryBuffer()
    m2.err_print_errors(err.bio_ptr())
    err_msg = err.read()
    if err_msg:
        return err_msg.decode()
    return None


def get_error_code() -> int:
    return m2.err_get_error()


def peek_error_code() -> int:
    return m2.err_peek_error()


def get_error_lib(err: Optional[int]) -> str:
    if err is None:
        return ""
    err_str = m2.err_lib_error_string(err)
    return err_str if err_str else ""


def get_error_func(err: Optional[int]) -> str:
    if err is None:
        return ""
    err_str = m2.err_func_error_string(err)
    return err_str if err_str else ""


def get_error_reason(err: Optional[int]) -> str:
    if err is None:
        return ""
    err_str = m2.err_reason_error_string(err)
    return err_str if err_str else ""


def get_error_message() -> str:
    return get_error_reason(get_error_code())


def get_x509_verify_error(err: Optional[int]) -> str:
    if err is None:
        return ""
    err_str = m2.x509_get_verify_error(err)
    return err_str if err_str else ""


class SSLError(Exception):
    def __init__(self, err: int, client_addr: util.AddrType) -> None:
        self.err = err
        self.client_addr = client_addr

    def __str__(self) -> str:
        if not isinstance(self.client_addr, str):
            s = (
                self.client_addr
                if isinstance(self.client_addr, str)
                else str(self.client_addr)
            )
        else:
            s = self.client_addr
        return "%s: %s: %s" % (
            get_error_func(self.err),
            s,
            get_error_reason(self.err),
        )


class M2CryptoError(Exception):
    pass



"""
M2Crypto wrapper for OpenSSL BN (BIGNUM) API.

Copyright (c) 2005 Open Source Applications Foundation. All rights reserved.
"""

from M2Crypto import m2
from typing import Optional  # noqa


def rand(bits: int, top: int = -1, bottom: int = 0) -> Optional[int]:
    """
    Generate cryptographically strong random number.

    :param bits:   Length of random number in bits.
    :param top:    If -1, the most significant bit can be 0. If 0, the most
                   significant bit is 1, and if 1, the two most significant
                   bits will be 1.
    :param bottom: If bottom is true, the number will be odd.
    """
    return m2.bn_rand(bits, top, bottom)


def rand_range(range: int) -> int:
    """
    Generate a random number in a range.

    :param range: Upper limit for range.
    :return:      A random number in the range [0, range)
    """
    return m2.bn_rand_range(range)


def randfname(length: int) -> str:
    """
    Return a random filename, which is simply a string where all
    the characters are from the set [a-zA-Z0-9].

    :param length: Length of filename to return.
    :return:       random filename string
    """
    import warnings

    warnings.warn(
        "Don't use BN.randfname(), use tempfile methods instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890"
    lettersLen = len(letters)
    fname: list = []
    for x in range(length):
        fname += [letters[m2.bn_rand_range(lettersLen)]]

    return "".join(fname)



# src/M2Crypto/types.py
"""
Type aliases for opaque C structures used in M2Crypto's C-bindings.
"""
from typing import Any, List, NewType

# The base type 'object' is used because these are opaque pointers
# at the Python level. NewType ensures they are treated as distinct types by mypy.
AES_KEY = NewType("AES_KEY", object)
ASN1_BitString = NewType("ASN1_BitString", object)
ASN1_Integer = NewType("ASN1_Integer", object)
ASN1_Object = NewType("ASN1_Object", object)
ASN1_String = NewType("ASN1_String", object)
ASN1_Time = NewType("ASN1_Time", object)
BIGNUM = NewType("BIGNUM", object)
BIO_METHOD = NewType("BIO_METHOD", object)
BIO = NewType("BIO", object)
DH = NewType("DH", object)
DSA = NewType("DSA", object)
DSA_SIG = NewType("DSA_SIG", object)
EC = NewType("EC", object)
EC_KEY = NewType("EC_KEY", object)
ECDSA_SIG = NewType("ECDSA_SIG", object)
ENGINE = NewType("ENGINE", object)
EVP_CIPHER_CTX = NewType("EVP_CIPHER_CTX", object)
EVP_CIPHER = NewType("EVP_CIPHER", object)
EVP_MD_CTX = NewType("EVP_MD_CTX", object)
EVP_MD = NewType("EVP_MD", object)
EVP_PKEY_CTX = NewType("EVP_PKEY_CTX", object)
EVP_PKEY = NewType("EVP_PKEY", object)
HMAC_CTX = NewType("HMAC_CTX", object)
PKCS7 = NewType("PKCS7", object)
RC4_KEY = NewType("RC4_KEY", object)
RSA = NewType("RSA", object)
SSL_CIPHER = NewType("SSL_CIPHER", object)
SSL_CTX = NewType("SSL_CTX", object)
SSL_METHOD = NewType("SSL_METHOD", object)
SSL = NewType("SSL", object)
SSL_SESSION = NewType("SSL_SESSION", object)
X509_CRL = NewType("X509_CRL", object)
X509_EXTENSION = NewType("X509_EXTENSION", object)
X509 = NewType("X509", object)
X509_NAME = NewType("X509_NAME", object)
X509_NAME_ENTRY = NewType("X509_NAME_ENTRY", object)
X509_REQ = NewType("X509_REQ", object)
X509_STORE = NewType("X509_STORE", object)
X509_STORE_CTX = NewType("X509_STORE_CTX", object)
X509V3_CTX = NewType("X509V3_CTX", object)

# Generic STACK_OF type for simplicity
STACK_OF = List[Any]
STACK_OF_X509 = List[X509]
STACK_OF_SSL_CIPHER = List[SSL_CIPHER]
STACK_OF_X509_EXTENSION = List[X509_EXTENSION]



"""
SSL peer certificate checking routines

Copyright (c) 2004-2007 Open Source Applications Foundation.
All rights reserved.

Copyright 2008 Heikki Toivonen. All rights reserved.
"""

__all__ = [
    "SSLVerificationError",
    "NoCertificate",
    "WrongCertificate",
    "WrongHost",
    "Checker",
]

import re
import socket

from M2Crypto import X509, m2  # noqa
from typing import Optional, Union  # noqa

try:
    from re import Pattern
except ImportError:
    from typing import Pattern


class SSLVerificationError(Exception):
    pass


class NoCertificate(SSLVerificationError):
    pass


class WrongCertificate(SSLVerificationError):
    pass


class WrongHost(SSLVerificationError):
    def __init__(
        self,
        expectedHost: str,
        actualHost: Union[str, bytes],
        fieldName: str = "commonName",
    ) -> None:
        """
        This exception will be raised if the certificate returned by the
        peer was issued for a different host than we tried to connect to.
        This could be due to a server misconfiguration or an active attack.

        :param expectedHost: The name of the host we expected to find in the
                             certificate.
        :param actualHost:   The name of the host we actually found in the
                             certificate.
        :param fieldName:    The field name where we noticed the error. This
                             should be either 'commonName' or 'subjectAltName'.
        """
        if fieldName not in ("commonName", "subjectAltName"):
            raise ValueError(
                "Unknown fieldName, should be either commonName " + "or subjectAltName"
            )

        SSLVerificationError.__init__(self)
        self.expectedHost = expectedHost
        self.actualHost = actualHost
        self.fieldName = fieldName

    def __str__(self) -> str:
        actual = (
            self.actualHost.decode()
            if isinstance(self.actualHost, bytes)
            else self.actualHost
        )
        return "Peer certificate %s does not match host, expected %s, got %s" % (
            self.fieldName,
            self.expectedHost,
            actual,
        )


class Checker:

    # COMPATIBILITY: re.Pattern is available only from Python 3.7+
    numericIpMatch: Pattern[str] = re.compile(r"^[0-9]+(\.[0-9]+)*$")

    def __init__(
        self,
        host: Optional[str] = None,
        peerCertHash: Optional[bytes] = None,
        peerCertDigest: str = "sha256",
    ) -> None:
        self.host = host
        self.fingerprint = peerCertHash
        self.digest: str = peerCertDigest

    def __call__(
        self, peerCert: Optional[X509.X509], host: Optional[str] = None
    ) -> bool:
        if peerCert is None:
            raise NoCertificate("peer did not return certificate")

        if host is not None:
            self.host = host

        if self.fingerprint:
            if self.digest not in ("sha256"):
                raise ValueError('unsupported digest "%s"' % self.digest)

            if self.digest == "sha256":
                expected_len = 64
            else:
                raise ValueError("Unexpected digest {0}".format(self.digest))

            if len(self.fingerprint) != expected_len:
                raise WrongCertificate(
                    (
                        "peer certificate fingerprint length does not match\n"
                        + "fingerprint: {0}\nexpected = {1}\n"
                        + "observed = {2}"
                    ).format(
                        self.fingerprint,
                        expected_len,
                        len(self.fingerprint),
                    )
                )

            expected_fingerprint = (
                self.fingerprint.decode()
                if isinstance(self.fingerprint, bytes)
                else self.fingerprint
            )
            observed_fingerprint = peerCert.get_fingerprint(md=self.digest)
            if observed_fingerprint != expected_fingerprint:
                raise WrongCertificate(
                    (
                        "peer certificate fingerprint does not match\n"
                        + "expected = {0},\n"
                        + "observed = {1}"
                    ).format(expected_fingerprint, observed_fingerprint)
                )

        if self.host:
            hostValidationPassed = False
            self.useSubjectAltNameOnly = False

            # subjectAltName=DNS:somehost[, ...]*
            try:
                subjectAltName = peerCert.get_ext("subjectAltName").get_value()
                if self._splitSubjectAltName(self.host, subjectAltName):
                    hostValidationPassed = True
                elif self.useSubjectAltNameOnly:
                    raise WrongHost(
                        expectedHost=self.host,
                        actualHost=subjectAltName,
                        fieldName="subjectAltName",
                    )
            except LookupError:
                pass

            # commonName=somehost[, ...]*
            if not hostValidationPassed:
                hasCommonName = False
                commonNames = ""
                for entry in peerCert.get_subject().get_entries_by_nid(
                    m2.NID_commonName
                ):
                    hasCommonName = True
                    commonName = entry.get_data().as_text()
                    if not commonNames:
                        commonNames = commonName
                    else:
                        commonNames += "," + commonName
                    if self._match(self.host, commonName):
                        hostValidationPassed = True
                        break

                if not hasCommonName:
                    raise WrongCertificate("no commonName in peer certificate")

                if not hostValidationPassed:
                    raise WrongHost(
                        expectedHost=self.host,
                        actualHost=commonNames,
                        fieldName="commonName",
                    )

        return True

    def _splitSubjectAltName(
        self,
        host: Union[str, bytes],
        subjectAltName: Union[str, bytes],
    ) -> bool:
        """
        >>> check = Checker()
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='DNS:my.example.com')
        True
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='DNS:*.example.com')
        True
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='DNS:m*.example.com')
        True
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='DNS:m*ample.com')
        False
        >>> check.useSubjectAltNameOnly
        True
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='DNS:m*ample.com, othername:<unsupported>')
        False
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='DNS:m*ample.com, DNS:my.example.org')
        False
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='DNS:m*ample.com, DNS:my.example.com')
        True
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='DNS:my.example.com, DNS:my.example.org')
        True
        >>> check.useSubjectAltNameOnly
        True
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='')
        False
        >>> check._splitSubjectAltName(host='my.example.com',
        ...     subjectAltName='othername:<unsupported>')
        False
        >>> check.useSubjectAltNameOnly
        False
        """
        host_str = host.decode() if isinstance(host, bytes) else host
        san_str = (
            subjectAltName.decode()
            if isinstance(subjectAltName, bytes)
            else subjectAltName
        )

        self.useSubjectAltNameOnly = False
        for certHost in san_str.split(","):
            certHost = certHost.lower().strip()
            if certHost[:4] == "dns:":
                self.useSubjectAltNameOnly = True
                if self._match(host_str, certHost[4:]):
                    return True
            elif certHost[:11] == "ip address:":
                self.useSubjectAltNameOnly = True
                if self._matchIPAddress(host_str, certHost[11:]):
                    return True
        return False

    def _match(self, host: str, certHost: str) -> bool:
        """
        >>> check = Checker()
        >>> check._match(host='my.example.com', certHost='my.example.com')
        True
        >>> check._match(host='my.example.com', certHost='*.example.com')
        True
        >>> check._match(host='my.example.com', certHost='m*.example.com')
        True
        >>> check._match(host='my.example.com', certHost='m*.EXAMPLE.com')
        True
        >>> check._match(host='my.example.com', certHost='m*ample.com')
        False
        >>> check._match(host='my.example.com', certHost='*.*.com')
        False
        >>> check._match(host='1.2.3.4', certHost='1.2.3.4')
        True
        >>> check._match(host='1.2.3.4', certHost='*.2.3.4')
        False
        >>> check._match(host='1234', certHost='1234')
        True
        """
        # XXX See RFC 2818 and 3280 for matching rules, this is may not
        # XXX yet be complete.

        host = host.lower()
        certHost = certHost.lower()

        if host == certHost:
            return True

        if certHost.count("*") > 1:
            # Not sure about this, but being conservative
            return False

        if self.numericIpMatch.match(host) or self.numericIpMatch.match(
            certHost.replace("*", "")
        ):
            # Not sure if * allowed in numeric IP, but think not.
            return False

        if certHost.find("\\") > -1:
            # Not sure about this, maybe some encoding might have these.
            # But being conservative for now, because regex below relies
            # on this.
            return False

        # Massage certHost so that it can be used in regex
        certHost = certHost.replace(".", "\\.")
        certHost = certHost.replace("*", "[^\\.]*")
        if re.compile("^%s$" % certHost).match(host):
            return True

        return False

    def _matchIPAddress(
        self, host: Union[str, bytes], certHost: Union[str, bytes]
    ) -> bool:
        """
        >>> check = Checker()
        >>> check._matchIPAddress(host='my.example.com',
        ...     certHost='my.example.com')
        False
        >>> check._matchIPAddress(host='1.2.3.4', certHost='1.2.3.4')
        True
        >>> check._matchIPAddress(host='1.2.3.4', certHost='*.2.3.4')
        False
        >>> check._matchIPAddress(host='1.2.3.4', certHost='1.2.3.40')
        False
        >>> check._matchIPAddress(host='::1', certHost='::1')
        True
        >>> check._matchIPAddress(host='::1', certHost='0:0:0:0:0:0:0:1')
        True
        >>> check._matchIPAddress(host='::1', certHost='::2')
        False
        """
        try:
            canonical = socket.getaddrinfo(
                host,
                0,
                0,
                socket.SOCK_STREAM,
                0,
                socket.AI_NUMERICHOST,
            )
            certCanonical = socket.getaddrinfo(
                certHost,
                0,
                0,
                socket.SOCK_STREAM,
                0,
                socket.AI_NUMERICHOST,
            )
        except:
            return False
        return canonical == certCanonical


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""Deprecated, use the util module instead.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

import warnings

warnings.warn("Use the util module instead", DeprecationWarning)

from M2Crypto.util import genparam_callback, passphrase_callback



"""M2Crypto enhancement to Python's urllib for handling
'https' url's.

FIXME: it is questionable whether we need this old-style module at all. urllib
(not urllib2) is in Python 3 support just as a legacy API.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

import base64
import warnings

from M2Crypto import SSL, httpslib

from urllib.response import addinfourl
from typing import Optional, Union  # noqa

from urllib.request import *  # noqa for other modules to import
from urllib.parse import *  # noqa for other modules to import
from urllib.error import *  # noqa for other modules to import

if "URLopener" in globals():

    def open_https(
        self: URLopener,
        url: Union[str, bytes],
        data: Optional[bytes] = None,
        ssl_context: Optional[SSL.Context] = None,
    ) -> addinfourl:
        """
        Open URL over the SSL connection.

        :param url: URL to be opened
        :param data: data for the POST request
        :param ssl_context: SSL.Context to be used
        :return:
        """
        warnings.warn("URLOpener has been deprecated in Py3k", DeprecationWarning)

        # Dynamically add the SSL context to the URLopener instance.
        context: SSL.Context
        if ssl_context is not None and isinstance(ssl_context, SSL.Context):
            context = ssl_context
        else:
            context = SSL.Context()

        self.ctx = context  # type: ignore[attr-defined]

        # Normalize the URL to a string for robust parsing.
        url_str = url.decode("latin-1") if isinstance(url, bytes) else url

        parsed = urlparse(url_str)
        if parsed.scheme.lower() != "https":
            raise IOError("url error", "invalid URL scheme")

        host = parsed.hostname
        if not host:
            raise IOError("http error", "no host given")

        # Build the host:port string for the connection.
        host_port = host
        if parsed.port:
            host_port += f":{parsed.port}"

        # Reconstruct the path and query part of the URL.
        selector = urlunsplit(("", "", parsed.path, parsed.query, parsed.fragment))
        if not selector:
            selector = "/"

        # Handle authentication from the URL.
        auth_header_value = None
        if parsed.username:
            user_pass = parsed.username
            if parsed.password:
                user_pass += f":{parsed.password}"
            # base64.encodebytes requires bytes and returns bytes.
            auth_bytes = base64.encodebytes(user_pass.encode("utf-8")).strip()
            # The final header value must be a string.
            auth_header_value = f'Basic {auth_bytes.decode("ascii")}'

        # Mypy gets confused by our re-defined HTTPSConnection.
        h = httpslib.HTTPSConnection(
            host=host_port, ssl_context=context
        )  # type: ignore[call-arg]

        if data is not None:
            h.putrequest("POST", selector)
            h.putheader("Content-type", "application/x-www-form-urlencoded")
            h.putheader("Content-length", str(len(data)))
        else:
            h.putrequest("GET", selector)

        if auth_header_value:
            h.putheader("Authorization", auth_header_value)

        # URLopener stores extra headers in `addheaders`.
        for header, value in self.addheaders:  # type: ignore[attr-defined]
            h.putheader(header, value)
        h.endheaders()

        if data is not None:
            h.send(data)

        # Get the response and wrap it in the expected addinfourl object.
        resp = h.getresponse()
        # Use the modern `.headers` attribute, not the deprecated `.msg`.
        return addinfourl(resp, resp.headers, "https:" + url_str)

    # Monkey-patch the method onto the (deprecated) URLopener class.
    URLopener.open_https = open_https  # type: ignore[attr-defined, method-assign, used-before-def, assignment]
else:
    import sys

    class URLopener:  # type: ignore [no-redef]
        msg = f'Python {"%d.%d" % (sys.version_info[:2])} does not support URLopener any more.'

        def __init__(self):
            raise RuntimeError(self.msg)

        def open_https(self):
            raise RuntimeError(self.msg)



"""M2Crypto enhancement to xmlrpclib.

Copyright (c) 1999-2003 Ng Pheng Siong. All rights reserved."""

import base64
import re

from M2Crypto import __version__ as __M2Version

from M2Crypto import SSL, httpslib, m2urllib2  # type: ignore
from typing import Callable, Optional, Union, Any, Dict, Tuple  # noqa

from xmlrpc.client import ProtocolError, Transport
from xmlrpc.client import *  # noqa


__version__ = __M2Version

HostType = Union[str, Tuple[str, Dict[str, str]]]
Marshallable = Any
SizedBufferType = Union[bytes, bytearray]


class SSL_Transport(Transport):

    user_agent = "M2Crypto_XMLRPC/%s - %s" % (
        __version__,
        Transport.user_agent,
    )

    def __init__(self, ssl_context: Optional[SSL.Context] = None, *args, **kw) -> None:
        Transport.__init__(self, *args, **kw)
        if ssl_context is None:
            self.ssl_ctx = SSL.Context()
        else:
            # FIX: Add assertion to make test_init_junk_ssl_context pass
            assert isinstance(ssl_context, SSL.Context), ssl_context
            self.ssl_ctx = ssl_context

    @staticmethod
    def splituser(host):
        """splituser('user[:passwd]@host[:port]') --> 'user[:passwd]', 'host[:port]'."""
        match = re.match("^(.*)@(.*)$", host)
        if match:
            return match.group(1, 2)
        return None, host

    @staticmethod
    def splitport(host):
        """splitport('host:port') --> 'host', 'port'."""
        match = re.match("^(.*):([0-9]*)$", host)
        if match:
            host, port = match.groups()
            if port:
                return host, port
        return host, None

    # This ignore is necessary because SizedBuffer is a private Protocol we cannot import
    # and satisfy fully at the argument level.
    def request(
        self,
        host: HostType,
        handler: str,
        request_body: SizedBufferType,  # type: ignore[override]
        verbose: int = 0,
    ) -> Tuple[Marshallable, ...]:

        # Handle username and password.
        user_passwd, host_port = self.splituser(host)
        _host, _port = self.splitport(host_port)

        _host_str = _host.decode("latin-1") if isinstance(_host, bytes) else _host

        h = httpslib.HTTPSConnection(_host_str, int(_port), ssl_context=self.ssl_ctx)  # type: ignore[call-arg]

        try:
            if verbose:
                h.set_debuglevel(1)

            h.putrequest("POST", handler)

            # required by HTTP/1.1
            h.putheader("Host", _host_str)

            # required by XML-RPC
            h.putheader("User-Agent", self.user_agent)
            h.putheader("Content-Type", "text/xml")
            h.putheader("Content-Length", str(len(request_body)))

            # Authorisation.
            if user_passwd is not None:
                auth = base64.encodebytes(user_passwd).strip()
                h.putheader("Authorization", "Basic %s" % auth.decode("ascii"))

            h.endheaders()

            if request_body:
                h.send(request_body)

            response = h.getresponse()

            host_str = host if isinstance(host, str) else str(host)

            headers_dict = dict(response.getheaders())

            if response.status != 200:
                raise ProtocolError(
                    host_str + handler,
                    response.status,
                    response.reason,
                    headers_dict,
                )

            self.verbose = verbose
            return self.parse_response(response)

        finally:
            h.close()
